import io
import json
import logging
import traceback
from copy import copy
from datetime import datetime
from types import TracebackType
from typing import Any, Optional, Tuple, Type, Union


class JsonFormatter(logging.Formatter):
    """
    Logging formatter for creating log entries in a JSON logstash-friendly format.
    Example:
        {
            "msg": ""GET / HTTP/1.1" 404 2327",
            "status_code": "404",
            "message": ""GET / HTTP/1.1" 404 2327",
            "server_time": "14/Sep/2016 11:13:00",
            "@timestamp": "2016-09-14 11:13:00,667"
        }
    """

    RESERVED_ATTRS = (
        "args",
        "asctime",
        "created",
        "exc_info",
        "exc_text",
        "filename",
        "funcName",
        "levelname",
        "levelno",
        "lineno",
        "module",
        "msecs",
        "message",
        "msg",
        "name",
        "pathname",
        "process",
        "processName",
        "relativeCreated",
        "stack_info",
        "thread",
        "threadName",
        "trace_id",
        "span_id",
        "dd_trace_id",
        "dd_span_id",
    )

    DEFAULT_FIELDS = (
        "asctime",
        "levelname",
        "filename",
        "funcName",
        "msg",
        "exc_info",
    )

    DEFAULT_MAPPING = {
        "asctime": "@timestamp",
    }

    converter = datetime.fromtimestamp  # type: ignore[assignment]
    default_time_format = "%Y-%m-%dT%H:%M:%S.%f"

    def __init__(
        self,
        *args: Any,
        rename: dict[str, str] | None = None,
        version: str = "1",
        fields: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Builds the formatter.
        param fmt: list or tuple containing default fields to include in every entry
        param datefmt: date format as a string to be passed to formatTime().
            Defaults to ISO8601 format.
        param rename: dictionary with {old_key: new_key} to be renamed in the log
            entries. Defaults to {'asctime':'@timestamp'}.
        param version: version as for @version field in logging, always included.
            Defaults to "1".
        """

        super().__init__(*args, **kwargs)

        self.fields = (
            [f for f in fields if f in self.RESERVED_ATTRS]
            if fields
            else self.RESERVED_ATTRS
        )

        self.rename_map = rename or self.DEFAULT_MAPPING
        self.version = version

    def formatTime(self, record: logging.LogRecord, datefmt: str | None = None) -> Any:
        """
        Overriding formatTime to add milliseconds parsing
        :param record:
        :param datefmt:
        :return:
        """

        ct = self.converter(record.created)  # type: ignore[has-type]
        _format = datefmt or self.default_time_format

        # noinspection PyUnresolvedReferences
        s = ct.strftime(_format)

        return s

    def format(self, record: logging.LogRecord) -> str:
        _msg = record.msg

        record.asctime = self.formatTime(record, self.datefmt)

        if isinstance(_msg, dict):
            msg_dict = _msg
        else:
            msg_dict = {}  # type: ignore[assignment]
            record.message = record.getMessage()

        extra_dict = {
            k: v
            for k, v in record.__dict__.items()
            if k not in self.RESERVED_ATTRS and not k.startswith("_")
        }

        # Fields specified at "fmt"
        fields_dict = {k: v for k, v in record.__dict__.items() if k in self.fields}

        # Adding fields coming from base message
        fields_dict.update(msg_dict)

        # Adding extra fields
        fields_dict.update(extra_dict)

        # Replacing fields names if rename mapping exists
        for k, v in self.rename_map.items():
            if k in fields_dict.keys():
                fields_dict[v] = fields_dict.pop(k)

        # Adding logging schema version if exists
        if self.version:
            fields_dict["@version"] = self.version

        return json.dumps(fields_dict, default=str)


class LogfmtFormatter(logging.Formatter):
    """
    Formatter that spits out logs parsable by logfmt in Loki. It has pretty sane
    defaults, but you can pass in what fields of the log record you want to expose. Any
    fields not found are just ignored. You pass in your custom fields as a list of
    tuples, that set the exposed name, and the field name in the record respectively.
    For example if you want to just expose the loglevel and call it "mycustomlevelname"
    in the logs you would do it like so:

        LogfmtFormatter([
            ("mycustomlevelname", "levelname"),
        ])
    """

    def __init__(self, fields: list[Tuple[str, str]] | None = None):
        super().__init__()
        if fields:
            self._fields = fields
        else:
            self._fields = [
                ("level", "levelname"),
                ("logger", "name"),
                ("http_client_addr", "http_client_addr"),
                ("http_method", "http_method"),
                ("http_path", "http_path"),
                ("http_version", "http_version"),
                ("http_status_code", "http_status_code"),
                ("duration", "duration"),
                ("trace_id", "trace_id"),
                ("span_id", "span_id"),
                ("dd_trace_id", "dd_trace_id"),
                ("dd_span_id", "dd_span_id"),
                ("msg", "msg"),
            ]

    def format(self, record: logging.LogRecord) -> str:
        record.asctime = datetime.fromtimestamp(record.created).isoformat()

        s = self.formatMessage(record)
        if record.exc_info:
            s += self.formatException(record.exc_info)
        if record.stack_info:
            s += self.formatStack(record.stack_info)
        return s

    def formatMessage(self, record: logging.LogRecord) -> str:
        s = f'time="{record.asctime}"'
        record_dict = record.__dict__
        for expose_key, record_key in self._fields:
            field = record_dict.get(record_key)
            if field:
                if record_key != "msg":
                    s += f' {expose_key}="{str(field)}"'
                else:
                    # noinspection PyBroadException
                    try:
                        msg = record.msg % record.args
                    except Exception:
                        msg = f"{record.msg} % {record.args}"
                    s += f' {expose_key}="{self._escape(msg)}"'
        return s

    def formatException(
        self,
        ei: Union[
            Tuple[Type[BaseException], BaseException, Optional[TracebackType]],
            Tuple[None, None, None],
        ],
    ) -> str:
        e_type, e_msg, e_trace = ei

        s = f' exception_type="{str(e_type)[8:-2]}"'
        s += f' exception_message="{self._escape(str(e_msg))}"'

        if e_trace:
            sio = io.StringIO()
            traceback.print_exception(e_type, e_msg, e_trace, None, sio)
            trace_s = sio.getvalue()
            sio.close()
            s += f' exception_trace="{self._escape(trace_s)}"'

        return s

    def formatStack(self, stack_info: str) -> str:
        return f' stack_info="{self._escape(stack_info)}"'

    @staticmethod
    def _escape(s: str) -> str:
        """
        Use json dumps to escape strings
        """

        escaped = json.dumps(s, ensure_ascii=False)
        return escaped[1 : len(escaped) - 1]  # noqa: E203


class PrettyFormatter(logging.Formatter):
    """
    Simple logging formatter for development and debugging
    """

    # This is inspired by the color logger in uvicorn
    _lvl_colors = {
        5: "\x1b[34m",  # Trace
        logging.DEBUG: "\x1b[36m",
        logging.INFO: "\x1b[32m",
        logging.WARNING: "\x1b[33m",
        logging.ERROR: "\x1b[31m",
        logging.CRITICAL: "\x1b[31m",
    }
    _color_reset = "\x1b[0m"

    def __init__(self, fmt: str | None = None, **kwargs: dict[str, Any]) -> None:
        super().__init__(
            fmt or "%(levelname)s %(message)s", **kwargs  # type: ignore[arg-type]
        )

    def formatMessage(self, record: logging.LogRecord) -> str:
        rc = copy(record)

        # Set level
        indent_message_start = 9
        separator = " " * (indent_message_start - len(rc.levelname))
        color = PrettyFormatter._lvl_colors.get(rc.levelno, "")
        reset = PrettyFormatter._color_reset if color != "" else ""
        rc.levelname = f"{color}{rc.levelname}{reset}:{separator}"

        # Make nice access logs if http attributes are present
        if (
            hasattr(rc, "http_client_addr")
            and hasattr(rc, "http_method")
            and hasattr(rc, "http_path")
            and hasattr(rc, "http_status_code")
            and hasattr(rc, "duration")
        ):
            rc.message = f"{rc.http_client_addr}"
            rc.message += " - "
            rc.message += f"{rc.http_method}"
            rc.message += f" {rc.http_path}"
            rc.message += f" {rc.http_status_code}"
            rc.message += f" {rc.duration}"
        if hasattr(rc, "trace_id"):
            rc.message += f" [trace_id: {rc.trace_id}]"
        if hasattr(rc, "dd_trace_id"):
            rc.message += f" [dd_trace_id: {rc.dd_trace_id}]"

        return super().formatMessage(rc)



from typing import Any


class DDLazy:
    """
    This class exists to make sure that ddtrace is not imported anywhere by troncos
    before tracing has been initialized.
    """

    def __init__(self) -> None:
        self._dd_trace_export_enabled = False
        self._dd_tracer = None
        self._dd_propagator = None

    def _on_loaded(self, export_dd_traces: bool) -> None:
        self._dd_trace_export_enabled = export_dd_traces
        self.dd_tracer()
        self.dd_propagator()

    def dd_initialized(self) -> bool:
        return self._dd_tracer is not None or self._dd_propagator is not None

    def dd_trace_export_enabled(self) -> bool:
        return self._dd_trace_export_enabled

    def dd_tracer(self) -> Any:
        if not self._dd_tracer:
            import ddtrace

            self._dd_tracer = ddtrace.tracer  # type: ignore[assignment]
        return self._dd_tracer

    def dd_propagator(self) -> Any:
        if not self._dd_propagator:
            import ddtrace.propagation.http

            self._dd_propagator = ddtrace.propagation.http.HTTPPropagator()  # type: ignore[assignment] # noqa: E501
        return self._dd_propagator


ddlazy = DDLazy()

__all__ = ["ddlazy"]



import logging
from typing import Literal, Tuple

from troncos.logs.filters import TraceIdFilter
from troncos.logs.formatters import JsonFormatter, LogfmtFormatter, PrettyFormatter

__all__ = [
    "JsonFormatter",
    "LogfmtFormatter",
    "PrettyFormatter",
    "init_logging_basic",
    "print_loggers",
]


def print_loggers(verbose: bool = True) -> None:
    """
    Function that prints out initialized loggers. This is helpful for you to visualize
    exactly how loggers have been set up in your project (and your dependencies). By
    default, all loggers will be printed. If you want to filter out logging
    placeholders, loggers with NullHandlers, and loggers that only propagate to parent,
    set the verbose parameter to False.

    This flowchart helps to debug logging issues:
    https://docs.python.org/3/howto/logging.html#logging-flow

    The output from this function will look something like this:

        Loggers:
        [ root                 ] logs.RootLogger LEVEL:0 PROPAGATE:True
          └ HANDLER logs.StreamHandler  LVL  20
            └ FILTER velodrome.observability.logs.TraceIdFilter
            └ FORMATTER velodrome.observability.logs.LogfmtFormatter
        [ uvicorn.access       ] logs.Logger LEVEL:20 PROPAGATE:False
        [ uvicorn.error        ] logs.Logger LEVEL:20 PROPAGATE:True
          └ FILTER velodrome.utils.obs._UvicornErrorFilter
        [ velodrome.access     ] logs.Logger LEVEL:20 PROPAGATE:True
          └ FILTER velodrome.observability.logs.HttpPathFilter
    """

    def internal(
        curr: Tuple[str, logging.Logger],
        rest: list[Tuple[str, logging.Logger]],
    ) -> None:
        i_name, i_log = curr

        print(
            f"[ {i_name.ljust(20)[:20]} ]"
            f" {str(i_log.__class__)[8:-2]}"
            f" LEVEL: {i_log.level if hasattr(i_log, 'level') else '?'}"
            f" PROPAGATE: {i_log.propagate if hasattr(i_log, 'propagate') else '?'}"
        )

        if hasattr(i_log, "filters"):
            for f in i_log.filters:
                print("  └ FILTER", str(f.__class__)[8:-2])

        if hasattr(i_log, "handlers"):
            for h in i_log.handlers:
                print(
                    "  └ HANDLER",
                    str(h.__class__)[8:-2],
                    " LEVEL:",
                    h.level if hasattr(h, "level") else "?",
                )
                if hasattr(h, "filters"):
                    for f in h.filters:
                        print("    └ FILTER", str(f.__class__)[8:-2])
                if hasattr(h, "formatter"):
                    print("    └ FORMATTER", str(h.formatter.__class__)[8:-2])

        if len(rest) > 0:
            curr = rest[0]
            rest = rest[1:]
            internal(curr, rest)

    all_but_root = []
    for name, logger in logging.Logger.manager.loggerDict.items():
        if not verbose:
            # Ignore placeholders
            if isinstance(logger, logging.PlaceHolder):
                continue

            # If it is a logger that does nothing but propagate to the parent, ignore
            if (
                len(logger.filters) == 0
                and len(logger.handlers) == 0
                and logger.propagate
            ):
                continue

            # If this logger only has the Null handler
            if (
                len(logger.filters) == 0
                and len(logger.handlers) == 1
                and isinstance(logger.handlers[0], logging.NullHandler)
            ):
                continue

        all_but_root.append((name, logger))

    all_but_root.sort()

    print("Loggers:")
    internal(("root", logging.getLogger()), all_but_root)  # type: ignore[arg-type]
    print("")


def init_logging_basic(
    *,
    level: str | int,
    formatter: Literal["cli", "logfmt", "json", "structlog"] | logging.Formatter,
) -> logging.Logger:
    """
    Setup root logger to handle trace_id in records.

    Loggers:
    [ root                 ] logging.RootLogger LEVEL: 20 PROPAGATE: True
      └ HANDLER logging.StreamHandler  LEVEL: 20
        └ FILTER troncos.logs.filters.TraceIdFilter
        └ FORMATTER troncos.logs.formatters.PrettyFormatter
    """

    # Create handler
    root_handler = logging.StreamHandler()
    root_handler.setLevel(level)
    root_handler.addFilter(TraceIdFilter())

    # Set formatter
    if formatter == "cli":
        root_handler.setFormatter(PrettyFormatter())
    elif formatter == "logfmt":
        root_handler.setFormatter(LogfmtFormatter())
    elif formatter == "json":
        root_handler.setFormatter(JsonFormatter())
    elif formatter == "structlog":
        # Formatter handled by structlog
        pass
    else:
        root_handler.setFormatter(formatter)

    # Configure root logger
    root = logging.getLogger()
    root.setLevel(level)
    root.handlers = [root_handler]

    return root



from troncos.traces.decorate import trace_block
from troncos.traces.propagation import activate_context_from_dict, get_trace_and_span_id

SPAN_ATTR_NAME = "_troncos_gunicorn_trace_span"


def pre_request_trace(worker, req):  # type: ignore[no-untyped-def]
    # Get potential context from headers
    req_headers = {}
    for k, v in req.headers:
        req_headers[str(k).lower()] = str(v)
    activate_context_from_dict(req_headers)

    span = trace_block("gunicorn.request", resource=__name__, span_type="web")
    setattr(worker, SPAN_ATTR_NAME, span)

    # Start span
    span.__enter__()


def post_request_trace(worker, _req, _environ, res):  # type: ignore[no-untyped-def]
    trace_id, _ = get_trace_and_span_id()
    span = getattr(worker, SPAN_ATTR_NAME, None)
    try:
        span.__exit__(None, None, None)  # type: ignore[union-attr]
    except:  # noqa: E722
        worker.log.exception("Exception finishing trace")
    return None if trace_id is None else f"{trace_id:x}"



import logging
from typing import Iterable, Literal

from structlog.contextvars import merge_contextvars

try:
    import structlog
    from structlog.types import Processor
except ImportError:
    raise Exception("This feature is only available if 'structlog' is installed")

from troncos.frameworks.structlog.processors import (
    StaticValue,
    trace_injection_processor,
)
from troncos.logs import init_logging_basic


def init_logging_structlog(
    level: str | int,
    formatter: Literal["cli", "logfmt"],
    app_release: str | None = None,
    environment: str | None = None,
    app_name: str | None = None,
    extra_processors: Iterable[Processor] | None = None,
) -> None:
    """
    Setting up Python logging with Structlog support

    :param level: Log level for logger and handler (minimum level for entries to be
        picked up/printed)
    :param formatter: The name of the formatter to use for log entries (must be defined
        inside this function)
    :param app_release: Optional string representing release version (i.e. "v22.1" or
        "0.0.1")
    :param environment: Optional string representing the environment (i.e. "prod" or
        "dev")
    :param app_name: Optional string with the name of the app/service (i.e. "trex" or
        "interno")
    :param extra_processors: Optional list of structlog processors to add to the chain
    """

    logger = init_logging_basic(
        level=level,
        formatter="structlog",
    )
    renderer: Processor

    # The renderer is the last processor in the chain of structlog processors, acts as
    # formatter
    if formatter == "cli":
        renderer = structlog.dev.ConsoleRenderer()
    elif formatter == "logfmt":
        renderer = structlog.processors.LogfmtRenderer()
    else:
        raise Exception("Invalid renderer configured")

    configure_structlog(
        renderer=renderer,
        release=app_release,
        environment=environment,
        app_name=app_name,
        handler=logger.handlers[0],
        extra_processors=extra_processors,
    )


def configure_structlog(
    renderer: Processor,
    handler: logging.Handler,
    release: str | None = None,
    environment: str | None = None,
    app_name: str | None = None,
    extra_processors: Iterable[Processor] | None = None,
) -> None:
    """

    :param renderer: Function which will be last in processor chain, will be used as
        formatter
    :param handler: The logging.Handler object used for the root logger
    :param release: Optional string representing release version (i.e. "v22.1" or
        "0.0.1")
    :param environment: Optional string representing the environment (i.e. "prod" or
        "dev")
    :param app_name: Optional string with the name of the app/service (i.e. "trex" or
        "interno")
    :param extra_processors: Optional list of structlog processors to add to the chain
    """

    # Use ISO-format for all timestamps output by structlog
    timestamper = structlog.processors.TimeStamper(fmt="ISO")

    # The shared processors are used both for entries originating from the builtin
    # Python logger, and entries that are created through structlog (
    # `structlog.get_logger()`)
    shared_processors: list[Processor] = [
        merge_contextvars,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        timestamper,
        structlog.processors.CallsiteParameterAdder(
            {
                structlog.processors.CallsiteParameter.FILENAME,
                structlog.processors.CallsiteParameter.FUNC_NAME,
                structlog.processors.CallsiteParameter.LINENO,
            }
        ),
        trace_injection_processor,
    ]

    # Append release string if that has been supplied in instantiation
    if release:
        shared_processors.append(StaticValue("release", release))

    # Append environment string if that has been supplied in instantiation
    if environment:
        shared_processors.append(StaticValue("environment", environment))

    # Append app name string if that has been supplied in instantiation
    if app_name:
        shared_processors.append(StaticValue("app_name", app_name))

    if extra_processors:
        shared_processors += extra_processors

    # Processing chain for log entries originating from outside structlog
    pre_chain = shared_processors + [
        structlog.stdlib.ExtraAdder(),
    ]

    # Processing chain for entries originating from structlog (from
    # `structlog.get_logger()`)
    _processors = shared_processors + [
        structlog.processors.StackInfoRenderer(),
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.format_exc_info,
    ]

    # All entries pass through structlog processors before going to the formatter
    structlog_processors = _processors + [
        structlog.stdlib.ProcessorFormatter.wrap_for_formatter
    ]

    # The formatter needs its own set of processors, ending with the renderer
    formatter_processors = [
        structlog.stdlib.ProcessorFormatter.remove_processors_meta,
        renderer,
    ]

    # The global configuration for structlog
    structlog.configure(
        processors=structlog_processors,
        logger_factory=structlog.stdlib.LoggerFactory(),
        # Caching makes this config immutable
        cache_logger_on_first_use=True,
    )

    # Setting the structlog formatter on our handler
    handler.setFormatter(
        structlog.stdlib.ProcessorFormatter(
            foreign_pre_chain=pre_chain, processors=formatter_processors
        )
    )



import logging
import os
from typing import Callable, Tuple

from troncos._ddlazy import ddlazy

logger = logging.getLogger(__name__)


def init_profiling_basic() -> Callable[[], Tuple[str, dict[str, str]]]:
    """
    Sets up and enables the continuous profiler.

    :return: A callable that returns the profile data and http headers that
             phlare is happy with.
    """

    if not ddlazy.dd_initialized():
        logger.warning(
            "If you are also tracing, YOU SHOULD CALL 'init_tracing_basic' "
            "BEFORE YOU INITIALIZE THE PROFILER"
        )

    # We import ddtrace here because we only want to import
    # it when this function is called, not when it is imported.
    import ddtrace
    from ddtrace.profiling.exporter import pprof  # type: ignore[attr-defined]

    # Define our custom exporter, again ve define this inside the function
    # so ddtrace does not get imported until the function is called.
    class _PprofExporter(pprof.PprofExporter):  # type: ignore[misc]
        pprof = ""

        @staticmethod
        def _map_frames(events):  # type: ignore[no-untyped-def]
            """
            This function takes the profiling events, and iterates over all the frames,
            formatting the data for phlare.
            """

            for _, e1 in events.items():
                for e2 in e1:
                    e2.frames = list(e2.frames)
                    for i in range(len(e2.frames)):
                        frame = e2.frames[i]
                        file = frame[0].split("/lib/", 1)[-1]
                        loc = f"{file}:{frame[1]}:{frame[2]}"
                        e2.frames[i] = (frame[0], frame[1], loc, frame[3])

        def export(self, events, start_time_ns, end_time_ns):  # type: ignore[no-untyped-def] # noqa: E501
            self._map_frames(events)  # type: ignore[no-untyped-call]
            pprof_profile, _ = super(_PprofExporter, self).export(
                events, start_time_ns, end_time_ns
            )
            self.pprof = pprof_profile.SerializeToString()

    _endpoint_exporter = _PprofExporter()

    @staticmethod  # type: ignore[misc]
    def _build_default_exporters():  # type: ignore[no-untyped-def]
        return [_endpoint_exporter]

    # Monkey patch default exporter func
    ddtrace.profiling.profiler._ProfilerInstance._build_default_exporters = _build_default_exporters  # type: ignore[assignment] # noqa: 501

    # Set params and enable continuous profiling
    os.environ.setdefault("DD_PROFILING_UPLOAD_INTERVAL", "14")
    import ddtrace.profiling.auto  # noqa: E402

    # Create function to return
    def python_pprof() -> Tuple[str, dict[str, str]]:
        """
        Returns the pprof content as a string and the headers that should be returned by
        a http endpoint
        """

        return _endpoint_exporter.pprof, {
            "Content-Type": "application/octet-stream",
            "Content-Disposition": 'attachment; filename="python.pprof"',
            "X-Content-Type-Options": "nosniff",
        }

    return python_pprof



"""
The processors are used by Structlog to process incoming log entries, a bit like
how the stdlib logging uses logging filters. This is currently implemented alongside
the filters to allow for parallel feature parity while we finish the current troncos
adoption.
"""
from troncos._ddlazy import ddlazy

try:
    from structlog.types import EventDict, WrappedLogger
except ImportError:
    raise Exception("This feature is only available if 'structlog' is installed")


class StaticValue:
    """
    Annotating log entries with values that are not subject to change after logger
    instantiation (i.e. version number or environment)

    :param key: The name of the variable
    :param value: The value of the variable

    Example:
        init_logging_structlog(
            ...,
            extra_processors=[
                StaticValue("process_start_time", str(datetime.now())),
                StaticValue("foo", "bar")
            ]
        )

        Would add the following data to the log entries,
            process_start_time=2022-10-31 14:27:11.081182 foo=bar
    """

    def __init__(self, key: str, value: str) -> None:
        self.key = key
        self.value = value

    def __call__(
        self, _logger: WrappedLogger, method: str, event_dict: EventDict
    ) -> EventDict:
        event_dict[self.key] = self.value
        return event_dict


def trace_injection_processor(
    _logger: WrappedLogger, _log_method: str, event_dict: EventDict
) -> EventDict:
    """
    Simple logging processor that adds a trace_id to the log record if available.
    """

    if not ddlazy.dd_initialized():
        return event_dict

    dd_context = ddlazy.dd_tracer().current_trace_context()
    if dd_context:
        event_dict["trace_id"] = f"{dd_context.trace_id:x}".zfill(32)
        event_dict["span_id"] = f"{dd_context.span_id:x}"

        if ddlazy.dd_trace_export_enabled():
            event_dict["dd_trace_id"] = dd_context.trace_id
            event_dict["dd_span_id"] = dd_context.span_id

    return event_dict



from typing import Any, Literal, Tuple

from troncos._ddlazy import ddlazy


def activate_context_from_dict(carrier: dict[str, str]) -> Any:
    """
    Extracts trace context from dict and activates it
    """

    context = get_context_from_dict(carrier)
    ddlazy.dd_tracer().context_provider.activate(context)


def get_context_from_dict(carrier: dict[str, str]) -> Any:
    """
    Gets trace context from a dictionary that contains propagation entries.
    """

    return ddlazy.dd_propagator().extract(carrier)


def add_context_to_dict(carrier: dict[str, str]) -> dict[str, str]:
    """
    Adds a trace "parent" entry to a dictionary. This injects all available formats
    """

    context = ddlazy.dd_tracer().current_trace_context()
    if context:
        ddlazy.dd_propagator().inject(context, carrier)
    return carrier


def get_propagation_value(fmt: Literal["w3c", "b3"] = "w3c") -> str | None:
    """
    Returns a trace propagation value
    """

    d: dict[str, str] = {}
    add_context_to_dict(d)
    if fmt == "b3":
        return d.get("b3")
    else:
        return d.get("traceparent")


def get_trace_and_span_id() -> Tuple[int | None, int | None]:
    """
    Returns either a tuple of (trace_id, span_id) of tuple of (None, None)
    """

    try:
        dd_context = ddlazy.dd_tracer().current_trace_context()
        return (dd_context.trace_id, dd_context.span_id)
    except:  # noqa: E722
        return (None, None)



import asyncio
import contextlib
import inspect
import logging

# noinspection PyUnresolvedReferences,PyProtectedMember
from functools import wraps
from types import FunctionType
from typing import (
    Any,
    Awaitable,
    Callable,
    Iterator,
    ParamSpec,
    Type,
    TypeVar,
    cast,
    overload,
)

from troncos._ddlazy import ddlazy

_TRACE_IGNORE_ATTR = "_trace_ignore"

TClass = TypeVar("TClass")

P = ParamSpec("P")
R = TypeVar("R")


@contextlib.contextmanager
def trace_block(
    name: str,
    *,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> Iterator[Any]:  # This is set to any, because we want to lazy-load ddtrace
    """
    Trace a code block using a with statement. Example:

    with trace_block("cool.block", resource="data!", attributes={"some": "attribute"}):
        time.sleep(1)
    """

    attributes = attributes or {}
    with ddlazy.dd_tracer().trace(
        name=name,
        resource=resource,
        service=service,
        span_type=span_type,
    ) as span:
        if attributes:
            span.set_tags(attributes)
        yield span


def _trace_function(
    f: Callable[P, R],
    name: str | None = None,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> Callable[P, R]:
    if inspect.iscoroutinefunction(f):
        # Async function
        @wraps(f)
        async def traced_func_async(*args: P.args, **kwargs: P.kwargs) -> R:
            with trace_block(
                name=name or f"{f.__module__}.{f.__qualname__}",
                resource=resource,
                service=service,
                span_type=span_type,
                attributes=attributes,
            ):
                awaitable_func = cast(Callable[P, Awaitable[R]], f)
                return await awaitable_func(*args, **kwargs)

        if hasattr(f, _TRACE_IGNORE_ATTR):
            return f  # type: ignore[return-value]

        return cast(Callable[P, R], traced_func_async)

    else:
        # "Regular" function
        @wraps(f)
        def traced_func(*args: P.args, **kwargs: P.kwargs) -> R:
            with trace_block(
                name=name or f"{f.__module__}.{f.__qualname__}",
                resource=resource,
                service=service,
                span_type=span_type,
                attributes=attributes,
            ):
                return f(*args, **kwargs)

        if hasattr(f, _TRACE_IGNORE_ATTR):
            return f

        return traced_func


@overload
def trace_function(
    fn: Callable[P, R],
    *,
    name: str | None = None,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> Callable[P, R]:
    ...


@overload
def trace_function(
    fn: None = None,
    *,
    name: str | None = None,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    ...


def trace_function(
    fn: Callable[P, R] | None = None,
    *,
    name: str | None = None,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> Callable[P, R] | Callable[[Callable[P, R]], Callable[P, R]]:
    """
    This decorator adds tracing to a function. Example:

    @trace_function
    def myfunc1()
        return "This will be traced"

    @trace_function(service="custom_service")
    def myfunc2()
        return "This will be traced as a custom service"
    """

    if fn and (callable(fn) or asyncio.iscoroutinefunction(fn)):
        assert fn  # type: ignore[truthy-function]
        return _trace_function(fn, name, resource, service, span_type, attributes)
    else:
        # No args
        def _inner(f: Callable[P, R]) -> Callable[P, R]:
            return _trace_function(f, name, resource, service, span_type, attributes)

        return _inner


@overload
def trace_class(
    klass: None = None,
    *,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> Callable[[Type[TClass]], Type[TClass]]:
    ...


@overload
def trace_class(
    klass: Type[TClass],
    *,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> Type[TClass]:
    ...


def trace_class(
    klass: Type[TClass] | None = None,
    *,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> Type[TClass] | Callable[[Type[TClass]], Type[TClass]]:
    """
    This decorator adds a tracing decorator to every method of the decorated class. If
    you don't want some methods to be traced, you can add the 'trace_ignore' decorator
    to them. Example:

    @trace_class
    class MyClass1:

        def m1(self):
            return "This will be traced"

        @trace_ignore
        def m2(self):
            return "This will not be traced"


    @trace_class(service="custom_service")
    class MyClass2:

        def m3(self):
            return "This will be traced as a custom service"
    """

    def _dec(cls: Type[TClass]) -> Type[TClass]:
        for key, value in cls.__dict__.items():
            if key.startswith("_"):
                continue
            if not (
                isinstance(value, FunctionType) or asyncio.iscoroutinefunction(value)
            ):
                continue

            logging.getLogger(__name__).debug(f"Tracing function {cls.__name__}.{key}")
            setattr(
                cls,
                key,
                _trace_function(
                    value,
                    name=None,
                    resource=resource,
                    service=service,
                    span_type=span_type,
                    attributes=attributes,
                ),
            )
        return cls

    if klass:
        return _dec(klass)
    else:
        return _dec


def trace_module(
    *,
    resource: str | None = None,
    service: str | None = None,
    span_type: str | None = None,
    attributes: dict[str, str] | None = None,
) -> None:
    """
    This function adds a tracing decorator to every function of the calling module. If
    you don't want some functions to be traced, you can add the 'trace_ignore' decorator
    to them. Example:

    # Start of module

    def my_function():
        return "This func will be traced"

    @trace_ignore
    def my_function():
        return "This func will not be traced"

    trace_module()

    # End of module
    """

    frame = inspect.stack()[1].frame
    scope = frame.f_locals
    module_name = scope.get("__name__", "unknown")
    for key, value in list(frame.f_locals.items()):
        if key.startswith("_"):
            continue
        if not (isinstance(value, FunctionType) or asyncio.iscoroutinefunction(value)):
            continue
        if getattr(value, "__module__", None) != module_name:
            continue

        logging.getLogger(__name__).debug(f"Tracing function {module_name}.{key}")
        scope[key] = _trace_function(
            value,
            name=None,
            resource=resource,
            service=service,
            span_type=span_type,
            attributes=attributes,
        )


def trace_ignore(f: Callable[P, R]) -> Callable[P, R]:
    """
    Decorator to disable automatic tracing of functions.
    See 'trace_module' and 'trace_class'.
    """

    setattr(f, _TRACE_IGNORE_ATTR, ())
    return f


def trace_set_span_attributes(attr: dict[str, str | bool | int]) -> None:
    """
    Add attributes to current span: Example:

    with trace_block("cool.block", attributes={"some": "attribute"}):
        result = calculate_something()
        trace_set_span_attributes({"result": result})
    """

    span = ddlazy.dd_tracer().current_span()
    if span:
        span.set_tags(attr)


def trace_set_traceback() -> None:
    ddlazy.dd_tracer().current_span().set_traceback()



import logging

from starlette.middleware import Middleware

from troncos.frameworks.asgi.middleware import AsgiLoggingMiddleware
from troncos.logs.filters import HttpPathFilter

try:
    from starlette.applications import Starlette
except ImportError:
    raise Exception("This feature is only available if 'starlette' is installed")


class _UvicornErrorFilter(logging.Filter):
    """
    This filter can be added to uvicorn logging so exceptions that happen when serving
    requests, are not logged 2 times. See 'init_uvicorn_logging' below for context.
    """

    def __init__(self, name: str = "UvicornErrorFilter") -> None:
        super().__init__(name)

    def filter(self, record: logging.LogRecord) -> bool:
        return record.msg != "Exception in ASGI application\n"


def init_uvicorn_logging(
    *,
    app: Starlette,
    logger_name: str | None = None,
    log_access_ignored_paths: list[str] | None = None,
) -> None:
    """
    This function sets up logging for uvicorn + starlette. There is a lot of meddling
    around with loggers in this function, so I will try my best to explain what I am
    trying to do here.

    First off, we cannot use the 'uvicorn.access' logger at all. The reason for that is
    because it logs messages after all middleware has been executed. Meaning that our
    tracing span is gone (like tears in the rain) when that logger logs. So we just
    disable 'uvicorn.access' and add new LoggingMiddleWare (see implementation in
    asgi/middleware.py) that handles logging, because that can see our tracing spans.
    The new loggers it uses are called 'velodrome.access' and 'velodrome.error'

    Now using our own middleware for logging poses another problem. Errors are logged
    twice! So, to mitigate that, we filter all exceptions that would be logged my our
    middleware out of the 'uvicorn.error' logger using _UvicornErrorFilter (above). We
    do not want to disable that logger completely, because it also logs other important
    messages.

    We also make sure that all loggers (except ROOT) just propagate records and none of
    them has a handler. Meaning all logs are funneled through ROOT. That is convenient,
    because it is a single place to add a "global" handler.

    Lastly we set the TraceIdFilter on the ROOT handler (yes, handler, not logger,
    because filters on loggers are not executed when record is propagated from a 'child'
    logger). We do this to add trace id to the records whenever they are available.

    It could help to take a look at this flowchart:
    https://docs.python.org/3/howto/logging.html#logging-flow

    So we end up with this situation in the end:
    [ root                 ] logging.RootLogger LEVEL: 20 PROPAGATE: True
      └ HANDLER logging.StreamHandler  LEVEL: 20
        └ FILTER troncos.logs.filters.TraceIdFilter
        └ FORMATTER troncos.logs.formatters.PrettyFormatter
    [ uvicorn.access       ] logging.Logger LEVEL: 20 PROPAGATE: False
    [ uvicorn.error        ] logging.Logger LEVEL: 20 PROPAGATE: True
      └ FILTER troncos.frameworks.starlette.uvicorn._UvicornErrorFilter
    [ velodrome.access     ] logging.Logger LEVEL: 20 PROPAGATE: True
      └ FILTER troncos.logs.filters.HttpPathFilter
    [ velodrome.error      ] logging.Logger LEVEL: 40 PROPAGATE:True
    """

    logger_name = logger_name or (
        getattr(app, "title") if hasattr(app, "title") else "starlette"
    )

    # Setup uvicorn by just propagating to root logger
    uvicorn = logging.getLogger("uvicorn")
    uvicorn.propagate = True
    uvicorn.handlers = []

    # Disable uvicorn.access log
    uvicorn_access = logging.getLogger("uvicorn.access")
    uvicorn.propagate = False
    uvicorn_access.handlers = []

    # Put a filter on uvicorn.error log
    uvicorn_error = logging.getLogger("uvicorn.error")
    uvicorn.propagate = True
    uvicorn_error.handlers = []
    uvicorn_error.addFilter(_UvicornErrorFilter())

    # Set up our custom access and error logger
    app_logger_access = logging.getLogger(f"{logger_name}.access")
    app_logger_access.propagate = True
    app_logger_access.addFilter(HttpPathFilter(log_access_ignored_paths))
    app_logger_access.setLevel(logging.INFO)

    app_logger_error = logging.getLogger(f"{logger_name}.error")
    app_logger_error.setLevel(logging.ERROR)

    # Add our middleware to starlette
    app.user_middleware.append(
        Middleware(AsgiLoggingMiddleware, logger_name=logger_name)
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['troncos',
 'troncos._ddlazy',
 'troncos.frameworks',
 'troncos.frameworks.asgi',
 'troncos.frameworks.gunicorn',
 'troncos.frameworks.starlette',
 'troncos.frameworks.structlog',
 'troncos.logs',
 'troncos.profiling',
 'troncos.traces']

package_data = \
{'': ['*']}

install_requires = \
['ddtrace==1.8.0', 'opentelemetry-sdk==1.16.0']

extras_require = \
{'grpc': ['opentelemetry-exporter-otlp-proto-grpc==1.16.0'],
 'http': ['opentelemetry-exporter-otlp-proto-http==1.16.0'],
 'structlog': ['structlog>=22.3.0,<23.0.0']}

setup_kwargs = {
    'name': 'troncos',
    'version': '3.1.1',
    'description': 'Collection of Python logging, tracing and profiling tools',
    'long_description': '<h1 align="center" style="border-bottom: 0">\n  🪵<br>\n  Troncos <br/>\n</h1>\n\n<p align="center">\n    <em>\n        Collection of Python logging, tracing and profiling tools\n    </em>\n    <br>\n    <a href="https://github.com/kolonialno/troncos/actions?workflow=CI">\n        <img src="https://github.com/kolonialno/troncos/actions/workflows/ci.yml/badge.svg" alt="CI status">\n    </a>\n    <a href="https://pypi.python.org/pypi/troncos">\n        <img src="https://img.shields.io/pypi/v/troncos.svg">\n    </a>\n    <img src="https://img.shields.io/pypi/pyversions/troncos">\n    <a href="https://github.com/kolonialno/troncos/blob/master/LICENSE">\n        <img src="https://img.shields.io/github/license/kolonialno/troncos.svg">\n    </a>\n</p>\n\n<!-- TOC -->\n  * [Etymology](#etymology)\n  * [Installation](#installation)\n  * [Setup](#setup)\n    * [Choosing a trace exporter](#choosing-a-trace-exporter)\n      * [gRPC](#grpc)\n      * [http](#http)\n    * [Import order](#import-order)\n    * [Plain setup](#plain-setup)\n    * [Starlette with uvicorn](#starlette-with-uvicorn)\n    * [Django](#django)\n  * [Logging](#logging)\n    * [Thoughts on logging](#thoughts-on-logging)\n    * [Log traces with Gunicorn](#log-traces-with-gunicorn)\n    * [Structlog](#structlog)\n  * [Tracing](#tracing)\n    * [Tracing your code](#tracing-your-code)\n      * [trace_function](#tracefunction)\n      * [trace_block](#traceblock)\n      * [trace_class](#traceclass)\n      * [trace_module](#tracemodule)\n      * [trace_ignore](#traceignore)\n    * [Trace Propagation](#trace-propagation)\n      * [Send context](#send-context)\n      * [Receive context](#receive-context)\n    * [Trace sampling](#trace-sampling)\n    * [Trace debugging](#trace-debugging)\n  * [Profiling](#profiling)\n    * [Setup endpoint](#setup-endpoint)\n    * [Enable scraping](#enable-scraping)\n  * [Development](#development)\n<!-- TOC -->\n\n## Etymology\n\n"Troncos" is the plural of the spanish word "Tronco", which translates to "trunk" or "log".\n\n## Installation\n\n```console\n# With pip\n$ pip install troncos\n\n# With poetry (grpc trace exporter)\n$ poetry add troncos -E grpc\n\n# With poetry (http trace exporter)\n$ poetry add troncos -E http\n```\n\n## Setup\n\n> **Note**: It is a good idea to use a `settings.py`-file (or similar) as an authoritative source of variables (service name, environment, whether tracing is enabled or not, log level, etc). In this README we mostly use `os.environ` for the sake of clarity.\n\n### Choosing a trace exporter\n\nTo export your traces, you have to pick either `grpc` or `http`. Using `grpc` gives you significant performance gains. If you are running a critical service with high load in production, we recommend using `grpc`.\n\n#### gRPC\n\n```toml\n[tool.poetry.dependencies]\ntroncos = {version="?", extras = ["grpc"]}\n```\n\n#### http\n\n```toml\n[tool.poetry.dependencies]\ntroncos = {version="?", extras = ["http"]}\n```\n\n> **Note**: You need to change the `TRACE_PORT` depending on your choice of protocol `http`/`grpc`.\n\n### Import order\n\nIt is very important that you do **NOT** import `ddtrace` anywhere before you have initialized troncos! Troncos should give you a warning if this is the case.\n\n### Plain setup\n\n```python\nfrom os import environ\n\nfrom troncos.logs import init_logging_basic\nfrom troncos.traces import init_tracing_basic, http_endpoint_from_env\nfrom troncos.profiling import init_profiling_basic\n\ninit_logging_basic(\n    level=environ.get("LOG_LEVEL", "INFO"),\n    formatter=environ.get("LOG_FORMATTER", "cli")  # Use "logfmt" or "json" in production\n)\ninit_tracing_basic(\n    endpoint=http_endpoint_from_env("TRACE_HOST", "TRACE_PORT", "/v1/traces"),\n    # endpoint_dd=http_endpoint_from_env("TRACE_DD_HOST", "TRACE_DD_PORT"),\n    service_name="my_service",\n    service_version=environ.get("VERSION", "unknown"),\n    service_env=environ.get("ENVIRONMENT", "localdev"),\n)\nprofiler = init_profiling_basic()\n\n# Import all your other stuff ...\n```\n\n### Starlette with uvicorn\n\n```python\nfrom os import environ\n\nfrom troncos.logs import init_logging_basic\nfrom troncos.traces import init_tracing_basic, http_endpoint_from_env\nfrom troncos.profiling import init_profiling_basic\n\ninit_logging_basic(\n    level=environ.get("LOG_LEVEL", "INFO"),\n    formatter=environ.get("LOG_FORMATTER", "cli")  # Use "logfmt" or "json" in production\n)\ninit_tracing_basic(\n    endpoint=http_endpoint_from_env("TRACE_HOST", "TRACE_PORT", "/v1/traces"),\n    # endpoint_dd=http_endpoint_from_env("TRACE_DD_HOST", "TRACE_DD_PORT"),\n    service_name="my_service",\n    service_version=environ.get("VERSION", "unknown"),\n    service_env=environ.get("ENVIRONMENT", "localdev"),\n)\nprofiler = init_profiling_basic()\n\n# Import all your other stuff ...\n\nfrom fastapi import FastAPI\nfrom troncos.frameworks.starlette.uvicorn import init_uvicorn_logging\n\napp = FastAPI(title="my_service")\ninit_uvicorn_logging(\n    app=app,\n    log_access_ignored_paths=["/health", "/metrics"],\n)\n```\n\n> **Note**: If you are running starlette but not calling `init_uvicorn_logging`, traces might not be logged.\n\n### Django\n\nAdd this logging and trace initialization to your `settings.py` file:\n\n```python\nimport environ\nfrom troncos.traces import init_tracing_basic\n\nenv = environ.Env()\n\nAPP_NAME = "my_service"\nVERSION = env.str("VERSION", default="unknown")\nENVIRONMENT = env.str("ENVIRONMENT", default="localhost")\n\n# ... All your settings here ...\n\n# Configure logging\n\nLOG_FORMATTER = env.str("LOG_FORMATTER", "logfmt")\nLOG_LEVEL = env.str("LOG_LEVEL", "INFO")\nLOGGING = {\n    "version": 1,\n    "disable_existing_loggers": True,\n    "filters": {\n        "trace_id": {"()": "troncos.logs.filters.TraceIdFilter"},\n    },\n    "formatters": {\n        "cli": {"()": "troncos.logs.formatters.PrettyFormatter"},\n        "json": {"()": "troncos.logs.formatters.JsonFormatter"},\n        "logfmt": {"()": "troncos.logs.formatters.LogfmtFormatter"},\n    },\n    "handlers": {\n        "console": {\n            "class": "logging.StreamHandler",\n            "formatter": LOG_FORMATTER,\n            "filters": ["trace_id"],\n        }\n    },\n    "loggers": {\n        APP_NAME: {"handlers": ["console"], "level": LOG_LEVEL},\n        "django": {"handlers": ["console"], "level": LOG_LEVEL},\n        "django.server": {\n            "handlers": ["console"],\n            "level": LOG_LEVEL,\n            "propagate": False,\n        },\n        "gunicorn.error": {\n            "handlers": ["console"],\n            "level": LOG_LEVEL,\n        },\n    },\n}\n\n# Configure tracing\n\nTRACING_ENABLED = env.bool("OPENTELEMETRY_TRACING_ENABLED", default=False)\nTRACING_HOST = env.str("OPENTELEMETRY_TRACING_HOST", default="localhost")\nTRACING_PORT = env.int("OPENTELEMETRY_TRACING_PORT", default=4318)\n\ninit_tracing_basic(\n    service_name=APP_NAME,\n    service_version=VERSION,\n    service_env=ENVIRONMENT,\n    endpoint=f"http://{TRACING_HOST}:{TRACING_PORT}/v1/traces"\n    if TRACING_ENABLED\n    else None,\n    # endpoint_dd=f"http://{TRACING_DD_HOST}:{TRACING_DD_PORT}"\n    # if TRACING_DD_ENABLED\n    # else None,\n)\n```\n\n> **Note**: This will not log traces of all incoming requests. See [log traces with Gunicorn](#log-traces-with-gunicorn) section on how to do that.\n\n## Logging\n\n### Thoughts on logging\n\nMore often then not, you want all loggers to propagate their records to the `root` logger and make the `root` logger handle everything. Depending on your project, this might require some additional configuration. Looking at the [python logging flow](https://docs.python.org/3/howto/logging.html#logging-flow) can help you understand how child loggers can propagate records to the `root` logger. Note that propagating to `root` is the default behaviour.\n\nThere is a nice helper function that will print all loggers in troncos called `print_loggers`:\n\n```python\nfrom troncos.logs import print_loggers\n\nprint_loggers(verbose=False)  # To visualize loggers\n```\n\nAfter calling the `init_logging_basic` function in a simple project you should see something like this printed `print_loggers`:\n\n```console\nLoggers:\n[ root                 ] logging.RootLogger LEVEL: 20 PROPAGATE: True\n  └ HANDLER logging.StreamHandler  LEVEL: 20\n    └ FILTER troncos.logs.filters.TraceIdFilter\n    └ FORMATTER troncos.logs.formatters.PrettyFormatter\n```\n\nSo in general, after the initial setup you can use any logger and that will propagate the log record to root:\n\n```python\nimport logging\n\nlogging.getLogger("my.random.logger").info("Root will handle this record")\n```\n\n### Log traces with Gunicorn\n\nCreate a `gunicorn/config.py` file in your project:\n\n```python\nimport time\n\ndef post_fork(server, worker):\n    pass\n\ndef pre_request(worker, req):\n    req._gunicorn_start_time = time.time()\n    worker.log.info("[begin] %s %s", req.method, req.path)\n    pass\n\ndef post_request(worker, req, environ, resp):\n    duration = time.time() - req._gunicorn_start_time\n    trace_id = next(iter([v for k, v in req.headers if k == "X-B3-TRACEID"]), None)\n    worker.log.info(\n        "[status=%s] %s %s duration=%s traceID=%s",\n        resp.status,\n        req.method,\n        req.path,\n        duration,\n        trace_id,\n    )\n```\n\nThen when running gunicorn, specify what config file to use:\n\n```console\ngunicorn myapp.wsgi:application --config python:myapp.gunicorn.config ...\n```\n\n### Structlog\n\nYou can substitute `init_logging_basic` with `init_logging_structlog` to setup structlog:\n\n```python\nfrom os import environ\nfrom troncos.frameworks.structlog.setup import init_logging_structlog\n\ninit_logging_structlog(\n    level=environ.get("LOG_LEVEL", "INFO"),\n    formatter=environ.get("LOG_FORMATTER", "cli"),  # Use "logfmt" in production\n)\n```\n\nAlternatively you can add trace injection into your own structlog setup:\n\n```python\nimport structlog\nfrom troncos.frameworks.structlog.processors import trace_injection_processor\n\nstructlog.configure(\n    processors=[\n       trace_injection_processor,\n    ],\n)\n```\n\n## Tracing\n\n### Tracing your code\n\nAfter initializing tracing in your project you can use different methods to trace your code.\n\n#### trace_function\n\nThis decorator adds tracing to a function. You can supply a tracer provider, if none is supplied, the global tracer provider will be used:\n\n```python\nfrom troncos.traces.decorate import trace_function\n\n@trace_function\ndef myfunc1():\n    return "This will be traced"\n\n@trace_function(service="my_custom_service")\ndef myfunc2():\n    return "This will be traced as my_custom_service"\n```\n\n#### trace_block\n\nTrace using a with statement. You can supply a tracer provider, if none is supplied, the global tracer provider will be used.\n\n```python\nfrom troncos.traces.decorate import trace_block\n\nwith trace_block(name="action", resource="thing", attributes={"some": "attribute"}):\n    print("... do an action to a thing...")\n```\n\n#### trace_class\n\nThis decorator adds a tracing decorator to every method of the decorated class. If you don\'t want some methods to be traced, you can add the [trace_ignore](#traceignore) decorator to them. You can supply a tracer provider, if none is supplied, the global tracer provider will be used:\n\n```python\nfrom troncos.traces.decorate import trace_class, trace_ignore\n\n@trace_class\nclass MyClass1:\n\n    def m1(self):\n        return "This will be traced"\n\n    @trace_ignore\n    def m2(self):\n        return "This will not traced"\n\n\n@trace_class(service="my_custom_service")\nclass MyClass2:\n\n    def m3(self):\n        return "This will be traced as my_custom_service"\n```\n\n#### trace_module\n\nThis function adds a tracing decorator to every function of the calling module. If you don\'t want some functions to be traced, you can add the [trace_ignore](#traceignore) decorator to them. You can supply a tracer provider, if none is supplied, the global tracer provider will be used:\n\n```python\nfrom troncos.traces.decorate import trace_ignore, trace_module\n\ndef my_function():\n    return "This func will be traced"\n\n@trace_ignore\ndef my_function():\n    return "This func will not be traced"\n\ntrace_module()\n```\n\n#### trace_ignore\n\nA decorator that will make [trace_class](#traceclass) and [trace_module](#tracemodule) ignore the decorated function/method.\n\n### Trace Propagation\n\n> **Warning**: Traces with IDs bigger than `64` bits are not propagated correctly because of limitations of [ddtrace](https://github.com/DataDog/dd-trace-py/blob/1e1de001d3fd694d3bcf0fff604a927ef891b19e/ddtrace/propagation/http.py#L102-L108). Default trace ID size for OTEL is `128` bits.\n\nIf you want to propagate your trace to the next service, you need to send/receive special headers with your request/message. If you are using plain `requests` that should be handled automatically by troncos. Here is how you do this manually:\n\n#### Send context\n\n```python\nfrom troncos.traces.propagation import get_propagation_value\n\n# Get header value\nvalue = get_propagation_value()\n\n# Send it somewhere\n```\n\nor\n\n```python\nfrom troncos.traces.propagation import add_context_to_dict\n\nsome_dict = {}\n\n# Add propagation headers to dict\nadd_context_to_dict(some_dict)\n\n# Send it somewhere\n```\n\n#### Receive context\n\nAgain, troncos should in most cases do this automatically for you, but here is how you do it manually:\n\n```python\nfrom troncos.traces.propagation import activate_context_from_dict\nfrom troncos.traces.decorate import trace_block\n\nsome_dict = {} \nactivate_context_from_dict(some_dict)\n\nwith trace_block("my_block"):\n    print("... do something ...")\n```\n\n### Trace sampling\n\nSet these variables to turn on trace sampling:\n\n```console\nOTEL_TRACES_SAMPLER=parentbased_traceidratio\nOTEL_TRACES_SAMPLER_ARG=0.05\nDD_TRACE_SAMPLE_RATE=0.05\n```\n\n### Trace debugging\n\nYou can enable trace debugging by setting the environmental variable `OTEL_TRACE_DEBUG=true`. That will print all spans to the console. If you would rather get the spans in a file you can also provide the variable `OTEL_TRACE_DEBUG_FILE=/tmp/traces`.\n\n## Profiling\n\n> **Warning**: Profiling while using Python 3.11 is [not yet fully supported](https://github.com/DataDog/dd-trace-py/issues/4149).\n\n### Setup endpoint\n\nSimply add a `/debug/pprof` endpoint that returns the profile:\n\n```python\nfrom fastapi import FastAPI\nfrom starlette.responses import Response\nfrom troncos.profiling import init_profiling_basic\n\nprofiler = init_profiling_basic()\n\napp = FastAPI(title="my_service")\n\n@app.get("/debug/pprof", response_model=str)\nasync def debug_pprof() -> Response:\n    content, headers = profiler()\n    return Response(content=content, headers=headers)\n```\n\nYou can verify that your setup works with the [pprof](https://github.com/google/pprof) cli:\n\n```console\n$ pprof -http :6060 "http://localhost:8080/debug/pprof"\n```\n\n> **Note**: You will get an empty string from `profiler()` until the first profile has been collected.\n\n### Enable scraping\n\nWhen you deploy your application, be sure to use the custom oda annotation for scraping:\n\n```yaml\nannotations:\n    phlare.oda.com/port: "8080"\n    phlare.oda.com/scrape: "true"\n```\n\n## Development\n\nWhen developing troncos you should be constantly aware of the fact that under no circumstances should you import `ddtrace` into your module. That would cause the tracer to initialize with default values that would not change when the user initializes the tracer.\n\nFor this reason, if you need to use any parts of `ddtrace` in your code, consider using the `_ddlazy` module, that gives you access lazily. If that does not satisfy your requirements you can always import `ddtrace` in your functions (not at the top of your module).\n\nYou can compare performance of your local version, to some older version of troncos using the [performance test setup](./perf).\n',
    'author': 'Guðmundur Björn Birkisson',
    'author_email': 'gudmundur.birkisson@oda.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kolonialno/troncos',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import logging

from troncos._ddlazy import ddlazy

logger = logging.getLogger(__name__)


class TraceIdFilter(logging.Filter):
    """
    Simple logging filter that adds a trace_id to the log record if available.

    It sounds strange to use filters to modify records, but this is described
    in the official python docs:

    https://docs.python.org/3/howto/logging-cookbook.html#using-filters-to-impart-contextual-information
    """

    def __init__(self, name: str = "TraceIdFilter") -> None:
        super().__init__(name)

    def filter(self, record: logging.LogRecord) -> bool:
        if not ddlazy.dd_initialized():
            return True

        dd_context = ddlazy.dd_tracer().current_trace_context()
        if dd_context:
            record.trace_id = f"{dd_context.trace_id:x}".zfill(32)
            record.span_id = f"{dd_context.span_id:x}"

            if ddlazy.dd_trace_export_enabled():
                record.dd_trace_id = dd_context.trace_id
                record.dd_span_id = dd_context.span_id

        return True


class HttpPathFilter(logging.Filter):
    """
    Simple logging filter that drops any log records regarding paths in the
    list 'ignored_paths'.
    """

    def __init__(
        self, ignored_paths: list[str] | None, name: str = "HttpPathFilter"
    ) -> None:
        super().__init__(name)
        self._ignored_paths = ignored_paths if ignored_paths else []

    def filter(self, record: logging.LogRecord) -> bool:
        if hasattr(record, "http_path"):
            return getattr(record, "http_path") not in self._ignored_paths
        return True


class ContextDetachExceptionDropFilter(logging.Filter):
    """
    There is a problem with OTEL where sometimes detaching context fails with an
    error. This issue is tracked here:

        https://github.com/open-telemetry/opentelemetry-python/issues/2606

    We have observed this in some services. In those cases the first (or root)
    span of an incoming starlette request fails to detach the context when it
    finishes. This does not seem to affect tracing, nor to cause any memory leaks.
    It just floods the logs with exceptions. So a "solution" can be to suppress
    those exceptions in the logs using this filter.

    So if you see this exception in your logs, you can consider using this filter:

    Traceback (most recent call last):
      File "opentelemetry/context/__init__.py", line 157, in detach
        _RUNTIME_CONTEXT.detach(token)  # type: ignore
        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
      File "opentelemetry/context/contextvars_context.py", line 50, in detach
        self._current_context.reset(token)  # type: ignore
        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    ValueError: <Token> was created in a different Context
    """

    def __init__(self, name: str = "ContextDetachExceptionDropFilter") -> None:
        super().__init__(name)
        self._count = 0
        logger.warning("ContextDetachExceptionDropFilter is no longer needed!")

    def filter(self, record: logging.LogRecord) -> bool:
        if (
            record.name == "opentelemetry.context"
            and record.msg == "Failed to detach context"
        ):
            if self._count == 0:
                logger.warning("Suppressing context detach exceptions")
            self._count += 1
            if self._count > 100:
                self._count = 0
            return False
        return True



from typing import Any

from opentelemetry.sdk.resources import Attributes, Resource
from opentelemetry.sdk.trace import Span, SpanContext, SpanProcessor
from opentelemetry.sdk.util.instrumentation import InstrumentationScope
from opentelemetry.trace import SpanKind, Status, StatusCode
from opentelemetry.trace.span import TraceFlags

from troncos import OTEL_LIBRARY_NAME, OTEL_LIBRARY_VERSION

_instrumentation_scope = InstrumentationScope(OTEL_LIBRARY_NAME, OTEL_LIBRARY_VERSION)
_default_trace_flags = TraceFlags(1)


def _internal_span_context(dd_span: Any) -> SpanContext:
    return SpanContext(
        dd_span.trace_id, dd_span.span_id, False, trace_flags=_default_trace_flags
    )


class _TranslatedSpan(Span):
    def __init__(
        self,
        dd_span: Any,
        base_resources: Attributes,
        default_resource: Resource,
        ignore_attrs: list[str],
    ) -> None:
        super().__init__(
            dd_span.name,
            _internal_span_context(dd_span),
            parent=self._create_parent_context(dd_span),
            sampler=None,
            trace_config=None,
            resource=self._create_resource(dd_span, base_resources, default_resource),
            kind=self._create_span_kind(dd_span),
            instrumentation_scope=_instrumentation_scope,
        )
        self.start(dd_span.start_ns)
        self._apply_translation(dd_span, ignore_attrs)
        self.end(dd_span.start_ns + dd_span.duration_ns)

    @staticmethod
    def _create_parent_context(dd_span: Any) -> SpanContext | None:
        if dd_span.parent_id:
            if not dd_span._parent:
                # External trace parent
                return SpanContext(dd_span.trace_id, dd_span.parent_id, True)
            else:
                return _internal_span_context(dd_span._parent)
        return None

    @staticmethod
    def _create_resource(
        dd_span: Any, base_attributes: Attributes, default_resource: Resource
    ) -> Resource:
        if default_resource.attributes["service.name"] == dd_span.service:
            return default_resource
        elif dd_span.service in ["fastapi", "flask", "starlette", "django"]:
            return default_resource

        # The resource constructor copies everything, so we just
        # set the service.name temporarily
        old_service = base_attributes["service.name"]
        base_attributes["service.name"] = dd_span.service
        res = Resource(base_attributes)
        base_attributes["service.name"] = old_service
        return res

    @staticmethod
    def _create_span_kind(dd_span: Any) -> SpanKind:
        kind = SpanKind.INTERNAL
        if dd_span.span_type:
            if dd_span.span_type in ["template"]:
                kind = SpanKind.INTERNAL
            elif dd_span.span_type in ["web"]:
                kind = SpanKind.SERVER
            elif dd_span.span_type in ["worker"]:
                kind = SpanKind.CONSUMER
            else:
                kind = SpanKind.CLIENT
        elif dd_span.name in ["celery.apply"]:
            kind = SpanKind.PRODUCER
        return kind

    def _apply_translation(self, dd_span: Any, ignore_attrs: list[str]) -> None:
        otel_error_attr_dict = {}
        otel_span_attr = {
            "resource": dd_span.resource,
        }

        if dd_span.span_type:
            otel_span_attr["dd_type"] = dd_span.span_type

        # Collect all "attributes" from the dd span
        dd_span_attr: dict[str, Any] = {
            **dd_span._meta,
            **dd_span._metrics,
        }
        dd_span_err_attr_mapping = {
            "error.msg": "exception.message",
            "error.type": "exception.type",
            "error.stack": "exception.stacktrace",
        }

        # Map set OTEL attributes based on DD attributes
        for k, v in dd_span_attr.items():
            otel_err_attr = dd_span_err_attr_mapping.get(k)
            if k.startswith("_dd"):
                continue
            elif otel_err_attr:
                otel_error_attr_dict[otel_err_attr] = v
            elif k not in ignore_attrs:
                otel_span_attr[k] = v

        self.set_attributes(otel_span_attr)

        # Map exception attributes
        if otel_error_attr_dict:
            # self.set_attributes(otel_error_attr_dict)  # Is this needed?
            self.add_event(name="exception", attributes=otel_error_attr_dict)

            status_exp_type = otel_error_attr_dict.get("exception.type", None)
            status_exp_msg = otel_error_attr_dict.get("exception.message", None)

            self.set_status(
                Status(
                    status_code=StatusCode.ERROR,
                    description=f"{status_exp_type}: {status_exp_msg}",
                )
            )


class DDSpanProcessor:
    """
    A ddtrace span processor that translates dd spans into otel span and
    sends them to otel span processors
    """

    def __init__(
        self,
        service_name: str,
        service_attributes: dict[str, str] | None,
        otel_span_processors: list[SpanProcessor],
        dd_traces_exported: bool = False,
        flush_on_shutdown: bool = True,
    ) -> None:
        self._otel_procs = otel_span_processors
        self._base_resources = {
            **(service_attributes or {}),
            **{"service.name": service_name},
        }
        self._default_resource = Resource(self._base_resources)  # type: ignore[arg-type] # noqa: 501
        self._dd_traces_exported = dd_traces_exported
        self._flush_on_shutdown = flush_on_shutdown
        self._dd_span_ignore_attr = [
            "runtime-id",
            "_sampling_priority_v1",
            "env",
            "version",
        ]
        for k in self._base_resources:
            self._dd_span_ignore_attr.append(k)

    def on_span_start(self, dd_span: Any) -> None:
        pass

    def on_span_finish(self, dd_span: Any) -> None:
        span = _TranslatedSpan(
            dd_span,
            base_resources=self._base_resources,  # type: ignore[arg-type]
            default_resource=self._default_resource,
            ignore_attrs=self._dd_span_ignore_attr,
        )
        if self._dd_traces_exported:
            span.set_attribute("dd_trace_id", str(dd_span.trace_id))
            span.set_attribute("dd_span_id", str(dd_span.span_id))
        for p in self._otel_procs:
            p.on_end(span)

    def shutdown(self, timeout: int) -> None:
        if self._flush_on_shutdown:
            for p in self._otel_procs:
                p.force_flush(timeout_millis=timeout)
                p.shutdown()



import logging
import time
from typing import Any, Awaitable, Callable

from troncos.frameworks.asgi.utils import guarantee_single_callable


class AsgiLoggingMiddleware:
    """
    ASGI application middleware that logs requests.
    """

    def __init__(
        self,
        app: Any,
        logger_name: str | None = None,
    ) -> None:
        self._app = guarantee_single_callable(app)
        ln = logger_name or "asgi"
        self._access = logging.getLogger(f"{ln}.access")
        self._error = logging.getLogger(f"{ln}.error")

    async def __call__(
        self,
        scope: dict[str, Any],
        receive: Callable[[Any], Any],
        send: Callable[[dict[str, Any]], Awaitable[Any]],
    ) -> Any:
        if scope["type"] != "http":
            return await self._app(scope, receive, send)

        client_ip, client_port = scope.get("client", ("NO_IP", -1))
        method = scope.get("method")
        path = scope.get("path")
        http_version = scope.get("http_version")
        status = [0]
        start_time = time.time()

        async def wrapped_send(message: dict[str, Any]) -> None:
            if "status" in message:
                status[0] = message.get("status", 0)

            await send(message)

            # "more_body" is used if there is still data to send. Log the request
            # when "http.response.body" has no more data left to send in the response.
            if message.get("type") == "http.response.body" and not message.get(
                "more_body", False
            ):
                self._access.info(
                    "",
                    extra={
                        "http_client_addr": f"{client_ip}:{client_port}",
                        "http_method": method,
                        "http_path": path,
                        "http_version": http_version,
                        "http_status_code": status[0],
                        "duration": f"{time.time()-start_time:.6f}",
                    },
                )

        try:
            return await self._app(scope, receive, wrapped_send)
        except Exception as e:
            self._error.error(
                "",
                exc_info=e,
                extra={
                    "http_client_addr": f"{client_ip}:{client_port}",
                    "http_method": method,
                    "http_path": path,
                    "http_version": http_version,
                    "http_status_code": 500,
                    "duration": f"{time.time()-start_time:.6f}",
                },
            )
            raise e



import logging
import os
import sys
from typing import Any

import opentelemetry
from opentelemetry.sdk.trace import SpanProcessor
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
    ConsoleSpanExporter,
    SimpleSpanProcessor,
)

from troncos._ddlazy import ddlazy
from troncos.traces.dd_shim import DDSpanProcessor

logger = logging.getLogger(__name__)
TRACE_HEADERS = [
    "accept",
    "cache-control",
    "content-security-policy",
    "content-type",
    "content-length",
    "expires",
    "location",
    "origin",
    "range",
    "referer",
    "retry-after",
    "server",
    "traceparent",
    "tracestate",
    "uber-trace-id",
    "x-b3-traceid",
    "x-country",
    "x-language",
    "x-xss-protection",
]


def http_endpoint_from_env(host_var: str, port_var: str, path: str = "") -> str | None:
    """
    Simple helper function construct http endpoint from environment variables
    """

    host = os.environ.get(host_var)
    if not host:
        return None
    port = os.environ.get(port_var)
    if not port:
        return None
    return f"http://{host}:{port}{path}"


def _create_span_processor(
    service_name: str,
    service_env: str | None = None,
    service_version: str | None = None,
    service_attributes: dict[str, str] | None = None,
    endpoint: str | SpanProcessor | None = None,
    endpoint_dd: str | None = None,
) -> DDSpanProcessor:
    """
    Creates a DD span processor that converts DD spans into OTEL spans
    """

    flush_on_shutdown = True
    otel_span_processors: list[SpanProcessor] = []
    if endpoint:
        if isinstance(endpoint, SpanProcessor):
            otel_span_processors.append(endpoint)
        else:
            # Create an exporter
            try:
                from opentelemetry.exporter.otlp.proto.grpc import (  # isort: skip # noqa: 501
                    trace_exporter,
                )

                logger.info("OTEL using GRPC exporter")
            except ImportError:  # pragma: no cover
                try:
                    from opentelemetry.exporter.otlp.proto.http import trace_exporter  # type: ignore[no-redef] # isort: skip # noqa: 501

                    logger.info("OTEL using HTTP exporter")
                except ImportError:  # pragma: no cover
                    trace_exporter = None  # type: ignore[assignment]

            if trace_exporter:
                logger.info(f"OTEL traces exported to {endpoint}")
                otel_span_processors.append(
                    BatchSpanProcessor(
                        trace_exporter.OTLPSpanExporter(endpoint=endpoint)
                    )
                )
    else:
        if _bool_from_string(os.environ.get("TRONCOS_REUSE_OTEL_PROCESSOR", "false")):
            tp = opentelemetry.trace.get_tracer_provider()
            if hasattr(tp, "_active_span_processor"):
                logger.info("Reusing OTEL span processor")
                otel_span_processors.append(tp._active_span_processor)
                flush_on_shutdown = False

    # Fallback to InMemorySpanExporter
    if len(otel_span_processors) == 0:
        logger.warning("No OTEL span processor configured")

    # Setup OTEL debug processor
    otel_trace_debug = _bool_from_string(os.environ.get("OTEL_TRACE_DEBUG", "false"))
    otel_trace_debug_file = os.environ.get("OTEL_TRACE_DEBUG_FILE")
    if otel_trace_debug:
        logger.info(f"OTEL debug processor to {otel_trace_debug_file or 'stdout'}")
        debug_out = (
            sys.stdout
            if not otel_trace_debug_file
            else open(otel_trace_debug_file, "w")
        )
        otel_span_processors.append(
            SimpleSpanProcessor(ConsoleSpanExporter(out=debug_out))
        )

    service_attributes = service_attributes or {}

    if service_env:
        service_attributes["environment"] = service_env

    if service_version:
        service_attributes["version"] = service_version

    return DDSpanProcessor(
        service_name,
        service_attributes,
        otel_span_processors,
        dd_traces_exported=endpoint_dd is not None,
        flush_on_shutdown=flush_on_shutdown,
    )


def _bool_from_string(s: str) -> bool:
    return s.lower() in ["1", "true", "yes"]


def _class_index(some_list: list[Any], k: Any) -> int | None:
    for index, thing in enumerate(some_list):
        if isinstance(thing, k):
            return index
    return None


def init_tracing_basic(
    service_name: str,
    service_env: str | None = None,
    service_version: str | None = None,
    service_attributes: dict[str, str] | None = None,
    endpoint: str | None = None,
    endpoint_dd: str | None = None,
    patch_modules: list[str] | None = None,
) -> None:
    """
    This function initializes tracing. It is important that it is call as early as
    possible in the execution of the program. It should be called before ddtrace
    is imported anywhere else.

    :param service_name:    Name of service ("myapi")
    :param service_env:     Optional service environment ("staging")
    :param service_version: Optional service version ("1.2.3")
    :param service_attributes: Optional service attributes
    :param endpoint:    Set this endpoint if you want to ship to Tempo
    :param endpoint_dd: Set this endpoint if you want to ship to DD
    :param patch_modules:   Optional list of modules that you want DD to patch, it
                            defaults to all modules
    """

    if ddlazy.dd_initialized():
        logger.warning("Function 'init_tracing_basic' called multiple times!")
        return

    if _bool_from_string(os.environ.get("TRONCOS_OMIT_ROOT_CONTEXT_DETACH", "false")):
        logger.warning("TRONCOS_OMIT_ROOT_CONTEXT_DETACH is no longer needed!")

    # Create custom span processor
    custom_dd_span_processor = _create_span_processor(
        service_name=service_name,
        service_env=service_env,
        service_version=service_version,
        service_attributes=service_attributes,
        endpoint=endpoint,
        endpoint_dd=endpoint_dd,
    )

    # Set service info
    service_env = service_env or "unset"
    service_version = service_version or "unset"
    os.environ.setdefault("DD_SERVICE", service_name)
    os.environ.setdefault("DD_ENV", service_env)
    os.environ.setdefault("DD_VERSION", service_version)

    # Set propagation info
    prop_extract_key = "DD_TRACE_PROPAGATION_STYLE_EXTRACT"
    prop_extract_val = "tracecontext,b3multi,b3 single header,datadog"
    os.environ.setdefault(prop_extract_key, prop_extract_val)

    prop_inject_key = "DD_TRACE_PROPAGATION_STYLE_INJECT"
    prop_inject_val = "tracecontext,b3 single header"
    os.environ.setdefault(prop_inject_key, prop_inject_val)

    # Disable telemetry and startup logs
    os.environ.setdefault("DD_INSTRUMENTATION_TELEMETRY_ENABLED", "false")
    os.environ.setdefault("DD_TRACE_STARTUP_LOGS", "false")

    # Setup dd endpoint
    if endpoint_dd:
        os.environ.setdefault("DD_TRACE_AGENT_URL", endpoint_dd)
        logger.info(f"DD traces exported to {endpoint_dd}")
    else:
        logger.info("DD traces not exported")

    ddtrace_already_imported = "ddtrace" in sys.modules

    # Initialize ddtrace
    import ddtrace

    # Shutdown default tracer immediately
    try:
        ddtrace.tracer.shutdown()
    except ValueError:
        pass

    class _PatchedDDTracer(ddtrace.Tracer):
        # Since we cannot patch _default_span_processors_factory,
        # we patch all functions that call it
        def __init__(self, *args, **kwargs):  # type: ignore[no-untyped-def]
            super().__init__(*args, **kwargs)
            self._fix_span_processors()

        def configure(self, *args, **kwargs):  # type: ignore[no-untyped-def]
            super().configure(*args, **kwargs)
            self._fix_span_processors()

        def _child_after_fork(self):  # type: ignore[no-untyped-def]
            super()._child_after_fork()  # type: ignore[no-untyped-call]
            self._fix_span_processors()

        def _fix_span_processors(self) -> None:
            processors = self._span_processors

            # Remove dd span aggregator if needed
            dd_span_aggregator = _class_index(
                processors, ddtrace.internal.processor.trace.SpanAggregator
            )
            if dd_span_aggregator and not endpoint_dd:
                processors.pop(dd_span_aggregator)

            # Add custom processor
            if not _class_index(processors, DDSpanProcessor):
                processors.append(custom_dd_span_processor)  # type: ignore[arg-type]

    # Patch dd tracer and create new tracer
    ddtrace.Tracer = _PatchedDDTracer  # type: ignore[misc]
    ddtrace.tracer = ddtrace.Tracer()  # type: ignore[no-untyped-call]

    # Check if someone imported ddtrace before us. We do this by checking if the
    # variables we set above were in fact used
    if ddtrace_already_imported:
        logger.warning(
            "DETECTED THAT ddtrace WAS IMPORTED BY ANOTHER MODULE BEFORE 'init_tracing_basic' WAS CALLED. THIS IS BAD!",  # noqa: E501
        )

        # Try to fix what we can in this situation

        if os.environ.get(prop_inject_key) == prop_inject_val:
            inject_set = set()
            inject_set.add(
                ddtrace.internal.constants._PROPAGATION_STYLE_W3C_TRACECONTEXT
            )
            inject_set.add(
                ddtrace.internal.constants.PROPAGATION_STYLE_B3_SINGLE_HEADER
            )
            ddtrace.config._propagation_style_inject = inject_set  # type: ignore[assignment] # noqa: E501

        if os.environ.get(prop_extract_key) == prop_extract_val:
            extract_set = ddtrace.internal.constants.PROPAGATION_STYLE_ALL
            ddtrace.config._propagation_style_extract = extract_set  # type: ignore[assignment] # noqa: E501

    # Configure what headers to trace
    ddtrace.config.trace_headers(TRACE_HEADERS)  # type: ignore[no-untyped-call]

    # Set dd tracer tags if attributes are set
    if service_attributes:
        ddtrace.tracer.set_tags(service_attributes)

    # Patch either all modules, or the ones specified by the user
    if patch_modules is None:
        ddtrace.patch_all()
    elif len(patch_modules) > 0:
        kwargs = {m: True for m in patch_modules}
        ddtrace.patch(**kwargs)  # type: ignore[arg-type]

    # Mark ddtrace as initialized in ddlazy
    ddlazy._on_loaded(endpoint_dd is not None)
